from enum import Enum
from typing import Any, Optional

from pypika import Field, Query, Table
from pypika.dialects import MSSQLQueryBuilder, PostgreSQLQueryBuilder, QueryBuilder
from pypika.enums import Dialects
from pypika.utils import builder, format_quotes

from metrics_layer.core.model.definitions import Definitions

PostgreSQLQueryBuilder.ALIAS_QUOTE_CHAR = None
PostgreSQLQueryBuilder.QUOTE_CHAR = None


class NullSorting(Enum):
    first = "FIRST"
    last = "LAST"


class QueryBuilderWithOrderByNullsOption(QueryBuilder):
    @builder
    def replace_table(self, current_table: Optional[Table], new_table: Optional[Table]) -> "QueryBuilder":
        """
        Replaces all occurrences of the specified table with the new table. Useful when reusing fields across
        queries.

        :param current_table:
            The table instance to be replaces.
        :param new_table:
            The table instance to replace with.
        :return:
            A copy of the query with the tables replaced.
        """
        self._from = [new_table if table == current_table else table for table in self._from]
        self._insert_table = new_table if self._insert_table == current_table else self._insert_table
        self._update_table = new_table if self._update_table == current_table else self._update_table

        self._with = [alias_query.replace_table(current_table, new_table) for alias_query in self._with]
        self._selects = [select.replace_table(current_table, new_table) for select in self._selects]
        self._columns = [column.replace_table(current_table, new_table) for column in self._columns]
        self._values = [
            [value.replace_table(current_table, new_table) for value in value_list]
            for value_list in self._values
        ]

        self._wheres = self._wheres.replace_table(current_table, new_table) if self._wheres else None
        self._prewheres = self._prewheres.replace_table(current_table, new_table) if self._prewheres else None
        self._groupbys = [groupby.replace_table(current_table, new_table) for groupby in self._groupbys]
        self._havings = self._havings.replace_table(current_table, new_table) if self._havings else None
        # Adding the slot for nulls first/last is the only change here
        self._orderbys = [
            (orderby[0].replace_table(current_table, new_table), orderby[1], orderby[2])
            for orderby in self._orderbys
        ]
        self._joins = [join.replace_table(current_table, new_table) for join in self._joins]

        if current_table in self._select_star_tables:
            self._select_star_tables.remove(current_table)
            self._select_star_tables.add(new_table)

    @builder
    def orderby(self, *fields: Any, **kwargs: Any) -> "QueryBuilder":
        for field in fields:
            field = Field(field, table=self._from[0]) if isinstance(field, str) else self.wrap_constant(field)

            self._orderbys.append((field, kwargs.get("order"), kwargs.get("nulls")))

    def _orderby_sql(
        self,
        quote_char: Optional[str] = None,
        alias_quote_char: Optional[str] = None,
        orderby_alias: bool = True,
        **kwargs: Any,
    ) -> str:
        """
        Produces the ORDER BY part of the query.  This is a list of fields and possibly their
        directionality, ASC or DESC and null sorting option (FIRST or LAST).
        The clauses are stored in the query under self._orderbys as a list of tuples
        containing the field, directionality (which can be None),
        and null sorting option (which can be None).

        If an order by field is used in the select clause,
        determined by a matching, and the orderby_alias
        is set True then the ORDER BY clause will use
        the alias, otherwise the field will be rendered as SQL.
        """
        clauses = []
        selected_aliases = {s.alias for s in self._selects}
        for field, directionality, nulls in self._orderbys:
            term = (
                format_quotes(field.alias, alias_quote_char or quote_char)
                if orderby_alias and field.alias and field.alias in selected_aliases
                else field.get_sql(quote_char=quote_char, alias_quote_char=alias_quote_char, **kwargs)
            )

            if directionality is not None:
                orient = f" {directionality.value}"
            else:
                orient = ""

            if nulls is not None:
                null_sorting = f" NULLS {nulls.value}"
            else:
                null_sorting = ""
            clauses.append(f"{term}{orient}{null_sorting}")

        return " ORDER BY {orderby}".format(orderby=",".join(clauses))


class SnowflakeQuery(Query):
    """
    Defines a query class for use with Snowflake.
    """

    @classmethod
    def _builder(cls, **kwargs) -> "SnowflakeQueryBuilderWithOrderByNullsOption":
        return SnowflakeQueryBuilderWithOrderByNullsOption(**kwargs)


class SnowflakeQueryBuilderWithOrderByNullsOption(QueryBuilderWithOrderByNullsOption):
    QUOTE_CHAR = None
    ALIAS_QUOTE_CHAR = None
    QUERY_ALIAS_QUOTE_CHAR = ""
    QUERY_CLS = SnowflakeQuery

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dialect=Dialects.SNOWFLAKE, **kwargs)


class PostgresQuery(Query):
    """
    Defines a query class for use with Snowflake.
    """

    @classmethod
    def _builder(cls, **kwargs) -> PostgreSQLQueryBuilder:
        return PostgreSQLQueryBuilder(**kwargs)


class PostgresQueryWithOrderByNullsOption(Query):
    """
    Defines a query class for use with Snowflake.
    """

    @classmethod
    def _builder(cls, **kwargs) -> "PostgreSQLQueryBuilderWithOrderByNullsOption":
        return PostgreSQLQueryBuilderWithOrderByNullsOption(**kwargs)


class PostgreSQLQueryBuilderWithOrderByNullsOption(
    PostgreSQLQueryBuilder, QueryBuilderWithOrderByNullsOption
):
    QUERY_CLS = PostgresQueryWithOrderByNullsOption


class RedshiftQuery(Query):
    """
    Defines a query class for use with Amazon Redshift.
    """

    @classmethod
    def _builder(cls, **kwargs) -> "RedShiftQueryBuilderWithOrderByNullsOption":
        return RedShiftQueryBuilderWithOrderByNullsOption(dialect=Dialects.REDSHIFT, **kwargs)


class RedShiftQueryBuilderWithOrderByNullsOption(QueryBuilderWithOrderByNullsOption):
    ALIAS_QUOTE_CHAR = None
    QUOTE_CHAR = None
    QUERY_CLS = RedshiftQuery


class MSSQLQueryBuilderCorrectLimit(MSSQLQueryBuilder):
    QUOTE_CHAR = None

    @builder
    def limit(self, limit: int):
        self._top = limit


class MSSSQLQuery(Query):
    """
    Defines a query class for use with Microsoft SQL Server (and other T-SQL flavors).
    """

    @classmethod
    def _builder(cls, **kwargs) -> MSSQLQueryBuilderCorrectLimit:
        return MSSQLQueryBuilderCorrectLimit(**kwargs)


query_lookup = {
    Definitions.snowflake: SnowflakeQuery,
    Definitions.bigquery: SnowflakeQuery,  # In terms of quoting, these are the same
    Definitions.redshift: RedshiftQuery,
    Definitions.postgres: PostgresQueryWithOrderByNullsOption,
    Definitions.druid: PostgresQuery,  # druid core query logic is postgres compatible, minus null sorting
    Definitions.duck_db: PostgresQueryWithOrderByNullsOption,  # duck db core query logic = postgres
    Definitions.databricks: PostgresQueryWithOrderByNullsOption,  # databricks core query logic = postgres
    Definitions.trino: PostgresQueryWithOrderByNullsOption,  # trino core query logic = postgres
    Definitions.sql_server: MSSSQLQuery,
    Definitions.azure_synapse: MSSSQLQuery,  # Azure Synapse is a T-SQL flavor
}

if_null_lookup = {
    Definitions.snowflake: "ifnull",
    Definitions.bigquery: "ifnull",
    Definitions.redshift: "nvl",
    Definitions.postgres: "coalesce",
    Definitions.databricks: "coalesce",
    Definitions.druid: "nvl",
    Definitions.duck_db: "coalesce",
    Definitions.trino: "coalesce",
    Definitions.sql_server: "isnull",
    Definitions.azure_synapse: "isnull",
}



class AccessDeniedOrDoesNotExistException(Exception):
    def __init__(self, message: str, object_name: str, object_type: str):
        self.message = message
        self.object_name = object_name
        self.object_type = object_type

    def __str__(self):
        return self.message


class QueryError(Exception):
    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return self.message


class JoinError(QueryError):
    pass



import json
from collections import defaultdict
from copy import copy
from itertools import combinations, product

import networkx

from metrics_layer.core.exceptions import (
    AccessDeniedOrDoesNotExistException,
    QueryError,
)
from metrics_layer.core.model.definitions import Definitions

from .base import SQLReplacement
from .join import Join, ZenlyticJoinRelationship, ZenlyticJoinType


class IdentifierTypes:
    primary = "primary"
    foreign = "foreign"
    join = "join"

    options = [primary, foreign, join]


class JoinGraph(SQLReplacement):
    def __init__(self, project) -> None:
        self.project = project
        self._join_preference = [
            ZenlyticJoinRelationship.one_to_one,
            ZenlyticJoinRelationship.many_to_one,
            ZenlyticJoinRelationship.one_to_many,
            ZenlyticJoinRelationship.many_to_many,
        ]
        self._merged_result_graph = None
        self._graph = None
        self._field_memo = {}
        self._weak_graph_memo = {}
        self._strong_graph_memo = {}

    def subgraph(self, view_names: list):
        return self.graph.subgraph(view_names)

    @property
    def graph(self):
        if self._graph is None:
            self._graph = self.build()
        return self._graph

    def list_join_graphs(self):
        graph = self.project.join_graph.graph
        sorted_components = self._strongly_connected_components(graph)
        return [f"subquery_{i}" for i, _ in enumerate(sorted_components)]

    def join_graph_hash(self, view_name: str) -> str:
        if view_name not in self._strong_graph_memo:
            graph = self.project.join_graph.graph
            sorted_components = self._strongly_connected_components(graph)

            sorted_comps = enumerate(sorted_components)
            graph_hash = next((f"subquery_{i}" for i, comps in sorted_comps if view_name in comps), None)
            if graph_hash is None:
                raise QueryError(
                    f"View name {view_name} not found in any joinable part of your data model. "
                    "Please make sure this is the right name for the view."
                )

            self._strong_graph_memo[view_name] = graph_hash
        return self._strong_graph_memo[view_name]

    def weak_join_graph_hashes(self, view_name: str) -> list:
        if view_name not in self._weak_graph_memo:
            graph = self.project.join_graph.graph
            sorted_components = self._strongly_connected_components(graph)

            join_graph_hashes = []
            for i, components in enumerate(sorted_components):
                subgraph_nodes = self._subgraph_nodes_from_components(graph, components)
                if view_name in subgraph_nodes:
                    join_graph_hashes.append(f"subquery_{i}")

            self._weak_graph_memo[view_name] = join_graph_hashes
        return self._weak_graph_memo[view_name]

    def get_joinable_view_names(self, view_name: str):
        graph = self.project.join_graph.graph
        sorted_components = self._strongly_connected_components(graph)

        joinable_views = []
        for components in sorted_components:
            subgraph_nodes = self._subgraph_nodes_from_components(graph, components)
            if view_name in subgraph_nodes:
                joinable_views.extend(list(subgraph_nodes))

        return [v for v in sorted(list(set(joinable_views))) if v != view_name]

    def _strongly_connected_components(self, graph):
        components = networkx.strongly_connected_components(graph)
        # Sort the sub-components graphs alphabetically
        sorted_sub_components = [list(sorted(c)) for c in components]
        # Sort by largest component, then inverse alphabetically based on the first view name
        sorted_components = sorted(sorted_sub_components, key=lambda x: (len(x), x[0]), reverse=True)
        return sorted_components

    @staticmethod
    def _subgraph_nodes_from_components(graph, components):
        edges = networkx.edge_dfs(graph, source=components)
        all_edges = list(edges) + [list(components)]
        return list(set(node for edge in all_edges for node in edge))

    def ordered_joins(self, view_pairs: list):
        joins = []
        joined_views = []
        for base_view, join_view in view_pairs:
            join = self.get_join(base_view, join_view)
            if join_view not in joined_views:
                joins.append(join)
            joined_views.append(join_view)
        return joins

    def collect_errors(self):
        errors = []
        for join in self.joins():
            errors.extend(join.collect_errors())
        return errors

    def joins(self):
        joins = []
        for base_view_name, join_view_name in self.graph.edges():
            joins.append(self.get_join(base_view_name, join_view_name))
        return joins

    def get_join(self, base_view_name: str, join_view_name: str):
        join_info = self.graph[base_view_name][join_view_name]
        join_definition = {**join_info, "base_view_name": base_view_name, "join_view_name": join_view_name}
        return Join(join_definition, project=self.project)

    def build(self):
        graph = networkx.DiGraph()
        identifier_map, primary_keys = self._identifier_map()
        self.composite_keys = self._composite_keys(primary_keys)
        reference_map = self._reference_map()
        views_seen = set()
        for view in self.project.views():
            if view.name in views_seen:
                raise QueryError(
                    f"Duplicate view names found in your project for the name {view.name}."
                    " Please make sure all view names are unique (note: join_as on identifiers "
                    "will create a view under its that name and the name must be unique)."
                )
            views_seen.add(view.name)
            graph.add_node(view.name)
            if view.name in reference_map:
                # Add all explicit "join" type references
                for join_view_name, join_identifier in reference_map[view.name].items():
                    graph.add_edge(view.name, join_view_name, **join_identifier)

            for identifier in view.identifiers:
                only_join = identifier.get("only_join", [])
                # Add all identifier matches across other views
                for join_view_name in identifier_map.get(identifier["name"], []):
                    if join_view_name != view.name and self._allowed_join(only_join, join_view_name):
                        join_view = self.project.get_view(join_view_name)
                        join_identifier = join_view.get_identifier(identifier["name"])
                        join_only_join = join_identifier.get("only_join", [])
                        if not self._allowed_join(join_only_join, view.name):
                            continue

                        join_info = self._identifier_to_join(
                            first_identifier=identifier,
                            first_view_name=view.name,
                            second_identifier=join_identifier,
                            second_view_name=join_view_name,
                        )
                        is_fanout = self._is_fanout(join_info["relationship"])
                        if is_fanout and join_view_name not in identifier.get("allowed_fanouts", []):
                            continue

                        # Make sure the new join is preferable to the old one
                        if graph.has_edge(view.name, join_view_name):
                            existing = graph[view.name][join_view_name]["relationship"]
                            existing_score = self._join_preference.index(existing)
                            new_score = self._join_preference.index(join_info["relationship"])
                            # Only if the new identifier gives us a more preferable join will we change
                            if existing_score > new_score:
                                graph.add_edge(view.name, join_view_name, **join_info)
                        else:
                            graph.add_edge(view.name, join_view_name, **join_info)

        # print(networkx.to_dict_of_dicts(graph))
        return graph

    def merged_results_graph(self, model):
        if self._merged_result_graph is None:
            self._merged_result_graph = self._build_merged_results_graph(model)
        return self._merged_result_graph

    def _build_merged_results_graph(self, model):
        with_dates = [
            field
            for field in self.project.fields(model=model)
            if field.canon_date and field.field_type == "measure"
        ]
        mappings = model.get_mappings(dimensions_only=True)

        # Merged result shared date and field mapping
        graph = networkx.DiGraph()

        existing_root_nodes, join_group_hashes = self._add_canon_dates_to_merged_result(
            graph, with_dates, join_root=Definitions.canon_date_join_graph_root
        )
        self._add_mappings_to_merged_result(
            graph,
            mappings,
            must_exist_in=list(join_group_hashes),
            root_nodes=existing_root_nodes,
            measures_only=True,
        )

        ordered_hashes = sorted(list(join_group_hashes))
        for join_group_hash_1, join_group_hash_2 in combinations(ordered_hashes, 2):
            join_root = join_group_hash_1 + "_" + join_group_hash_2

            pair = [join_group_hash_1, join_group_hash_2]
            existing_sub_root_nodes, _ = self._add_canon_dates_to_merged_result(
                graph, with_dates, join_root, use_condition=True, must_be_in=pair
            )

            # Add any fields that accessible via a join to both join_group_hash_1 and join_group_hash_2
            for view in self.project.views(model=model):
                join_hashes = self.project.join_graph.weak_join_graph_hashes(view.name)
                if all(join_hash in join_hashes for join_hash in pair):
                    view_fields = self.project.fields(
                        view_name=view.name, model=model, expand_dimension_groups=True
                    )
                    for field in view_fields:
                        for node in existing_sub_root_nodes:
                            graph.add_edge(node, field.id())

            self._add_mappings_to_merged_result(
                graph, mappings, must_exist_in=pair, root_nodes=existing_sub_root_nodes
            )

        return graph

    def _add_canon_dates_to_merged_result(
        self, graph, measures: list, join_root: str, use_condition: bool = False, must_be_in: list = []
    ):
        self._field_memo = {}
        existing_root_nodes, join_group_hashes = set(), set()
        for measure in measures:
            join_hash = self.project.join_graph.join_graph_hash(measure.view.name)
            join_group_hashes.add(join_hash)
            if not use_condition or (use_condition and join_hash in must_be_in):
                measure_id = measure.id()
                try:
                    canon_date = self._get_field_with_memo(measure.canon_date, by_name=True)
                    for timeframe in canon_date.timeframes:
                        canon_date.dimension_group = timeframe
                        root_node_name = join_root + "_" + timeframe
                        graph.add_edges_from(
                            [(root_node_name, canon_date.id()), (root_node_name, measure_id)]
                        )
                        existing_root_nodes.add(root_node_name)
                except AccessDeniedOrDoesNotExistException:
                    # In the event that the canon_date doesn't exist anymore, don't break everything
                    pass
        return sorted(list(existing_root_nodes)), join_group_hashes

    def _add_mappings_to_merged_result(
        self, graph, mappings: dict, must_exist_in: list, root_nodes: list, measures_only: bool = False
    ):
        for from_field, mapping in mappings.items():
            if mapping.get("is_canon_date_mapping"):
                continue
            if measures_only and mapping["field_type"] != "measure":
                continue
            for reference in mapping["references"]:
                to_field = reference["field"]
                if mapping["from_join_hash"] in must_exist_in and reference["to_join_hash"] in must_exist_in:
                    from_ = self._get_field_with_memo(from_field).id()
                    to_ = self._get_field_with_memo(to_field).id()
                    for node in root_nodes:
                        graph.add_edges_from([(node, from_), (node, to_)])

    def _get_field_with_memo(self, field_name: str, by_name: bool = False):
        if field_name not in self._field_memo:
            if by_name:
                field = self.project.get_field_by_name(field_name)
            else:
                field = self.project.get_field(field_name)
            self._field_memo[field_name] = field
        else:
            field = self._field_memo[field_name]
        return field

    def _identifier_map(self):
        result = defaultdict(list)
        primary_keys = defaultdict(list)
        for view in self.project.views():
            for identifier in view.identifiers:
                if identifier["type"] != IdentifierTypes.join:
                    result[identifier["name"]].append(view.name)
                    # Make an additional mapping from identifier name to view's where it's a primary key
                    if identifier["type"] == IdentifierTypes.primary:
                        primary_keys[identifier["name"]].append(view.name)

        return result, primary_keys

    def _composite_keys(self, primary_key_map: dict):
        composite_keys = {}
        for view in self.project.views():
            for identifier in view.identifiers:
                # For composite identifiers, we need to add the individual
                # fields to indicate that they're part of the composite key
                if identifier["type"] != IdentifierTypes.join and "identifiers" in identifier:
                    composite_key_views = []
                    for sub_identifier in identifier["identifiers"]:
                        composite_key_views.append(primary_key_map.get(sub_identifier["name"], []))

                    for view_group in product(*composite_key_views):
                        composite_keys[tuple(sorted(view_group))] = view.name
        return composite_keys

    def _reference_map(self):
        result = defaultdict(dict)
        for view in self.project.views():
            for identifier in view.identifiers:
                if identifier["type"] == IdentifierTypes.join:
                    join_identifier = json.loads(json.dumps(identifier))
                    join_identifier["relationship"] = self._invert_relationship(
                        join_identifier["relationship"]
                    )
                    # We need to invert the join here because this is the inverse
                    # direction of how the join was defined
                    result[view.name][identifier["reference"]] = self._verify_identifier_join(identifier)

                    # We only want to invert the join by default if it's *not* a fanout join
                    if not self._is_fanout(join_identifier["relationship"]):
                        result[identifier["reference"]][view.name] = self._verify_identifier_join(
                            join_identifier
                        )

        return result

    def _identifier_to_join(self, first_identifier, first_view_name, second_identifier, second_view_name):
        relationship = self._derive_relationship(first_identifier, second_identifier)
        join_type = ZenlyticJoinType.left_outer
        first_clause = self._identifier_join_clause(first_identifier, first_view_name)
        second_clause = self._identifier_join_clause(second_identifier, second_view_name)
        sql_on = f"{first_clause}={second_clause}"
        weight = self._edge_weight(relationship)
        return {"relationship": relationship, "type": join_type, "sql_on": sql_on, "weight": weight}

    def _identifier_join_clause(self, identifier: dict, view_name: str):
        if "sql" in identifier:
            cleaned_sql = copy(str(identifier["sql"]))
            for field_name in self.fields_to_replace(str(identifier["sql"])):
                to_replace = "${" + field_name + "}"
                if field_name != "TABLE" and "." not in field_name:
                    cleaned_reference = "${" + f"{view_name}.{field_name}" + "}"
                    cleaned_sql = cleaned_sql.replace(to_replace, cleaned_reference)
                if field_name == "TABLE":
                    cleaned_sql = cleaned_sql.replace(to_replace, view_name)
            clause = cleaned_sql
        else:
            clause = "${" + f"{view_name}.{identifier['name']}" + "}"
        return clause

    def _verify_identifier_join(self, join: dict):
        clean_join = json.loads(json.dumps(join))
        clean_join["type"] = join.get("type", ZenlyticJoinType.left_outer)
        clean_join["relationship"] = join.get("relationship", ZenlyticJoinRelationship.many_to_one)
        clean_join["sql_on"] = join["sql_on"]
        clean_join["weight"] = self._edge_weight(clean_join["relationship"])
        return clean_join

    @staticmethod
    def _derive_relationship(identifier, join_identifier):
        base_type = identifier["type"]
        join_type = join_identifier["type"]
        if base_type == IdentifierTypes.foreign and join_type == IdentifierTypes.primary:
            return ZenlyticJoinRelationship.many_to_one
        elif base_type == IdentifierTypes.primary and join_type == IdentifierTypes.primary:
            return ZenlyticJoinRelationship.one_to_one
        elif base_type == IdentifierTypes.primary and join_type == IdentifierTypes.foreign:
            return ZenlyticJoinRelationship.one_to_many
        elif base_type == IdentifierTypes.foreign and join_type == IdentifierTypes.foreign:
            return ZenlyticJoinRelationship.many_to_many
        else:
            raise QueryError(
                "This join type cannot be determined from the identifier properties. "
                f"Make sure you've set the properties correctly. Base type: {base_type},"
                f" join type: {join_type}"
            )

    @staticmethod
    def _invert_relationship(relationship: str):
        mapping = {
            ZenlyticJoinRelationship.many_to_one: ZenlyticJoinRelationship.one_to_many,
            ZenlyticJoinRelationship.one_to_many: ZenlyticJoinRelationship.many_to_one,
            ZenlyticJoinRelationship.many_to_many: ZenlyticJoinRelationship.many_to_many,
            ZenlyticJoinRelationship.one_to_one: ZenlyticJoinRelationship.one_to_one,
        }
        return mapping[relationship]

    @staticmethod
    def _edge_weight(relationship: str):
        mapping = {
            ZenlyticJoinRelationship.many_to_one: 2,
            ZenlyticJoinRelationship.one_to_many: 3,
            ZenlyticJoinRelationship.many_to_many: 4,
            ZenlyticJoinRelationship.one_to_one: 1,
        }
        return mapping[relationship]

    @staticmethod
    def _allowed_join(only_join: list, view_name: str):
        return not only_join or (only_join and view_name in only_join)

    @staticmethod
    def _is_fanout(relationship: str):
        return relationship in {ZenlyticJoinRelationship.one_to_many, ZenlyticJoinRelationship.many_to_many}



import random
import string
import uuid


def generate_uuid(db_safe=False):
    if db_safe:
        return generate_random_password(40)
    return str(uuid.uuid4())


def generate_random_password(length):
    # Random string with the combination of lower and upper case
    letters = string.ascii_letters
    result_str = "".join(random.choice(letters) for i in range(length))
    return result_str


def flatten_filters(filters: list):
    flat_list = []

    def recurse(filter_obj):
        if isinstance(filter_obj, dict):
            if "conditions" in filter_obj:
                for f in filter_obj["conditions"]:
                    recurse(f)
            else:
                flat_list.append(filter_obj)
        elif isinstance(filter_obj, list):
            for item in filter_obj:
                recurse(item)

    recurse(filters)
    return flat_list
